import { modelSymbol, parentSymbol, apc, vpc, modelName } from './symbols';
import { assign } from './util';
import { emitter as mitt } from './emitter';
var modelMap = new Map();
var objKeys = Object.keys.bind(Object);
function diff(oldObj, newObj, whitelist) {
    if (oldObj === newObj)
        return;
    objKeys(oldObj)
        .reduce(function (acc, val) {
        return ~acc.indexOf(val) ? acc : acc.concat(val);
    }, objKeys(newObj))
        .forEach(function (key) {
        if (~whitelist.indexOf(key))
            return;
        var oldVal = oldObj[key];
        var newVal = newObj[key];
        if (oldVal === newVal)
            return;
        if (oldVal != null &&
            newVal != null &&
            typeof oldVal === 'object' &&
            typeof newVal === 'object') {
            oldVal[modelSymbol]
                ? oldVal.applySnapshot(newVal, true)
                : diff(oldVal, newVal, whitelist);
        }
        else {
            oldObj[apc][key] = newVal;
        }
    });
}
function setUpObject(obj, emit, symbol, path, parent) {
    if (parent === void 0) { parent = obj; }
    if (obj[modelName] && !obj[modelSymbol]) {
        obj = modelMap.has(obj[modelName])
            ? modelMap.get(obj[modelName])(obj)
            : obj;
    }
    if (obj[modelSymbol] && obj !== parent) {
        obj[parentSymbol](emit, parent, path);
        return obj;
    }
    else {
        // have to ignore errors because typescript doesn't like indexing with symbols
        if (!obj[apc] ||
            // @ts-ignore
            !obj[apc][symbol])
            obj[apc] = cProxy(obj, emit, symbol, false, path, parent);
        if (!obj[vpc] ||
            // @ts-ignore
            !obj[vpc][symbol])
            obj[vpc] = cProxy(obj, emit, symbol, true, path, parent);
    }
    for (var prop in obj) {
        var val = obj[prop];
        if (val != null && typeof val === 'object') {
            obj[prop] = setUpObject(val, emit, symbol, path + '/' + prop, parent);
        }
    }
    return obj;
}
function cProxy(obj, emit, symbol, view, path, parent) {
    return new Proxy(obj, {
        get: function (target, prop) {
            if (prop === symbol)
                return true;
            // @ts-ignore
            if (typeof prop === 'symbol')
                return target[prop];
            var res = target[prop];
            // check if there are other built in constructors that need this
            if (target instanceof Date && typeof res === 'function')
                return res.bind(target);
            return ((res != null &&
                typeof res === 'object' &&
                (view ? res[vpc] : res[apc])) ||
                res);
        },
        set: function (target, prop, val) {
            if (view) {
                throw new Error("You can't modify the state outside of actions");
            }
            if (target[prop] === val)
                return true;
            if (typeof prop === 'symbol') {
                target[prop] = val;
                return true;
            }
            if (val != null && typeof val === 'object') {
                val = setUpObject(val, emit, symbol, path + '/' + prop, parent || target[vpc]);
            }
            target[prop] = val;
            emit('patch', { path: path + '/' + prop, op: 'replace', value: val });
            return true;
        }
    });
}
function subscribe(emitter, evt, fn) {
    emitter.on(evt, fn);
    return function () { return emitter.off(evt, fn); };
}
export function model(name, _a) {
    var initial = _a.initial, actions = _a.actions, views = _a.views;
    if (typeof initial !== 'function') {
        throw new Error('You have to supply a function that returns the initial state');
    }
    if (actions != null && typeof actions !== 'function') {
        throw new TypeError('actions has to be a function');
    }
    if (views != null && typeof views !== 'function') {
        throw new TypeError('views has to be a function');
    }
    function instantiate(obj) {
        if (obj === void 0) { obj = {}; }
        var symbol = Symbol('model');
        var path = '';
        var parentEmit = null;
        var parent = null;
        var emitter = mitt();
        var state = assign(initial(), obj);
        var emit = function (evt, val) {
            if (evt === 'snapshot') {
                val = common.getSnapshot();
            }
            else if (val.path != null) {
                val.path = path + val.path;
            }
            emitter.emit(evt, val);
            parentEmit && parentEmit(evt, val);
        };
        var whitelist = [];
        var common = {
            onAction: function (fn, after) {
                var evt = after ? 'action-complete' : 'action';
                return subscribe(emitter, evt, fn);
            },
            onSnapshot: function (fn) {
                return subscribe(emitter, 'snapshot', fn);
            },
            onPatch: function (fn) {
                return subscribe(emitter, 'patch', fn);
            },
            getSnapshot: function () {
                return JSON.parse(JSON.stringify(state));
            },
            applySnapshot: function (snapshot, dontemit) {
                diff(state, snapshot, whitelist);
                !dontemit && emit('snapshot', {});
            },
            getParent: function () {
                return parent;
            },
            getRoot: function () {
                return (parent
                    ? parent.getRoot()
                        ? parent.getRoot()
                        : parent
                    : state);
            }
        };
        whitelist.push.apply(whitelist, objKeys(common));
        assign(state, common);
        state[modelSymbol] = true;
        state[parentSymbol] = function (emit, _parent, _path) {
            parentEmit = emit;
            path = _path;
            parent = _parent;
        };
        state[modelName] = name;
        state = setUpObject(state, emit, symbol, '');
        if (actions) {
            var boundActions = actions(state[apc]);
            var keys = objKeys(boundActions);
            var emitSnapshot_1 = function (name, args) {
                emit('action-complete', { name: name, path: '', args: args });
                emit('snapshot', {});
            };
            var _loop_1 = function (key) {
                var action = boundActions[key];
                state[key] = function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    emit('action', { name: key, path: '', args: args });
                    var res = action.apply(null, args);
                    if (res != null && res.then) {
                        res.then(emitSnapshot_1.bind(null, key, args));
                    }
                    else {
                        emitSnapshot_1(key, args);
                    }
                    return res;
                };
            };
            for (var _i = 0, keys_1 = keys; _i < keys_1.length; _i++) {
                var key = keys_1[_i];
                _loop_1(key);
            }
            whitelist.push.apply(whitelist, keys);
        }
        if (views) {
            var boundViews_1 = views(state[vpc]);
            var keys_3 = objKeys(boundViews_1);
            // r = refresh on get, v = value
            // I don't trust uglify here
            var caches_1 = {};
            var _loop_2 = function (key) {
                Object.defineProperty(state, key, {
                    get: function () {
                        var cache = caches_1[key];
                        if (cache.r) {
                            cache.r = false;
                            cache.v = boundViews_1[key]();
                        }
                        return cache.v;
                    },
                    configurable: false
                });
                caches_1[key] = { v: boundViews_1[key](), r: false };
            };
            for (var _a = 0, keys_2 = keys_3; _a < keys_2.length; _a++) {
                var key = keys_2[_a];
                _loop_2(key);
            }
            // update the cache on change
            state.onPatch(function () {
                for (var _i = 0, keys_4 = keys_3; _i < keys_4.length; _i++) {
                    var key = keys_4[_i];
                    caches_1[key].r = true;
                }
            });
        }
        // typeof state.init === 'function' && state.init();
        return state;
    }
    modelMap.set(name, instantiate);
    return function (obj) { return instantiate(obj)[vpc]; };
}
export function clearCache() {
    modelMap.clear();
}
//# sourceMappingURL=index.js.map